#include <string.h>
#include "mtsolver.h"
#include "global.h"
#include "globmat.h"
#include "mechprint.h"
#include "matrix.h"
#include "vector.h"
#include "gmatrix.h"
#include "gtopology.h"
#include "dloadcase.h"

#ifndef FNAMELEN
 #define FNAMELEN 1001
#endif

/**
   function solves time dependent mechanical problems with negligible inertial forces
   examples of such problems: creep, viscoplasticity
   
   JK
*/
void solve_time_dep_prob ()
{
  long lcid;
  
  //  load case must be equal to zero, no superposition can be used
  //  in this type of analysis
  lcid=0;
  
  //  solver of the problem
  visco_solver (lcid);
  
}


/**
   function solves time dependent mechanical problem (creep, viscoplasticity)
   
   @param lcid - load case id
   
   27.10.2001
*/
void visco_solver (long lcid)
{
  long i,j,k,n,ini,nsts;
  double dt,dtmin,dtdef,end_time,zero,*f,*fl,*fi,*fb,*fp,*r,*dr,*lhsb;
  double norf, norfa, err;
  matrix lsm_a(3,3);
  vector lsm_r(3), lsm_l(3);
  
  /*
  Mp->filename_decomposition (Outdm->outfn,path,name,suffix);
  sprintf (fname,"%s%s.bac",path, name);

  delete [] path; delete [] name; delete [] suffix;

  if (Mp->timecon.nit > 0)
    aux = fopen (fname,"w");
  */

  //  maximum number of iterations in inner loop
  ini = Mp->niilnr;
  //  required norm of vector of unbalanced forces
  err = Mp->errnr;
  //  number of rows of the matrix
  n = Ndofm;
  //  computer zero
  zero = Mp->zero;
  //  minimum time increment
  dtmin=Mp->timecon.dtmin;

  fp = new double [n];
  fb = new double [n];
  fl = new double [n];
  fi = new double [n];
  dr = new double [n];
  lhsb = new double [n];
  memset (fp,0,n*sizeof(double));
  memset (fb,0,n*sizeof(double));
  memset (fl,0,n*sizeof(double));
  memset (fi,0,n*sizeof(double));
  memset (dr,0,n*sizeof(double));
  memset (lhsb,0,n*sizeof(double));
  
  //  initialization phase
  r = Lsrs->give_lhs (lcid);
  f = Lsrs->give_rhs (lcid);
  
  //  initial time
  Mp->time=Mp->timecon.starttime ();
  //  end time
  end_time = Mp->timecon.endtime ();
  //  initial time increment
  dt = Mp->timecon.initialtimeincr ();
  
  //  allocation of backup arrays for state variables in integration points
  //Mm->alloc_backup ();
  
  //  provizorni umisteni, bude v mefelinit
  Gtm->comp_domain_sizes ();
  Mb->alloc_sumcomp ();
  //  konec provizoria

  //  initiation of mechanical material models
  Mm->initmaterialmodels();

  nullv (fp,n);
  i=0;
  //  number of successful time steps (number of steps without inner iterations)
  nsts=0;
  
  print_init(-1, "wt");
  print_step(lcid, i, Mp->time, f);
  print_flush();
  //print_close();
  
  // ***************************
  //   main iteration loop  ****
  // ***************************
  do{
    //new time increment
    Mp->time = Mp->timecon.newtime (dt);
    //  new step number
    i++;
    //  indicator of strain computation
    //  it means, strains at integration points have not been computed
    Mp->strainstate=0;
    //  indicator of stress computation
    //  it means, stresses at integration points have not been computed
    Mp->stressstate=0;
    //  indicator of computation of other array
    //  it means, stresses at integration points have not been computed
    Mp->otherstate=0;

    //  backup of attained nodal values
    for (j=0;j<n;j++){
      lhsb[j]=r[j];
    }
    
    if (Mespr==1)  fprintf (stdout,"\n\n -------------------------------------------------------------");
    if (Mespr==1)  fprintf (stdout,"\n Time step = %ld,  Time %e,  Time increment = %e",i,Mp->time,dt);
    if (Mespr==1)  fprintf (stdout,"\n -------------------------------------------------------------\n");

    //  prescribed nodal forces
    mefel_right_hand_side (lcid,f,fl);

    Mb->comp_sum (f);
    fprintf(Out, "\n# i=%ld, time=%le[s], time=%le[d], fx=%le, fy=%le, fz=%le", 
            i, Mp->time, Mp->time/86400, Mb->sumcomp[0], Mb->sumcomp[1], Mb->sumcomp[2]);
    fflush(Out);
    
    //  increments of prescribed forces
    incr_internal_forces (lcid,fi);
    
    //  right hand side assembling
    for (j=0;j<n;j++){
      fb[j]=f[j]-fp[j]+fi[j];
    }
    
    //  assembling of tangent stiffness matrix
    stiffness_matrix (lcid);
    
    //  solution of K(r).v=F
    Mp->ssle->solve_system (Gtm,Smat,dr,fb,Out);
    
    //  new displacement vector
    for (j=0;j<n;j++){
      r[j]+=dr[j];
    }
    
    //  computation of internal forces
    internal_forces (lcid,fi);
        
    norfa=ss(f,f,n);
    for (j=0;j<n;j++){
      fb[j] = fl[j] - fi[j];
    }
    //  norm of vector of unbalanced forces
    norf=ss(fb,fb,n);
    if (norfa > 0.0)
      norf /= norfa;

    fprintf (stdout,"\n\n Norf before inner iteration loop norf = %e\n\n\n",norf);

    j = 0;
    if (norf<err){
      nsts++;      
    }
    else{
      // iteration of unbalanced forces caused by time independent models
      //  internal iteration loop
      fillm(0.0, lsm_a);
      fillv(0.0, lsm_r);
      for (j=0;j<ini;j++)
      {
        //  solution of K(r).v=F
	//Smat->solve_system (Gtm,dr,fb);
	Mp->ssle->solve_system (Gtm,Smat,dr,fb,Out);
	  
	for (k=0;k<n;k++)
	  r[k]+=dr[k];
	  
	//  computation of internal forces
	internal_forces (lcid,fi);
	  
	//  vector of unbalanced forces
	for (k=0;k<n;k++)
	  fb[k]=fl[k] - fi[k];
	  
	//  norm of vector of unbalanced forces
	norf=ss(fb,fb,n);
	if (norfa > 0.0)
	  norf /= norfa;
	  
	if (Mespr==1)
//        if ((Mespr==1)&&(j%10==0)) // print each 10-th step
  	  fprintf (stdout,"\n Inner loop number j = %ld,     norf = %e \n",j,norf);
	  
	if (norf<err)  break;

        // divergence detection with help of least square method
        if (j > 1)
        {
          if (lsm_quad(lsm_a, lsm_r, lsm_l, double(j+1), norf, Mp->zero,1) > 0.0)
	  {
	    fprintf (stdout,"\n\n Divergence of inner iteation loop is detected. Inner loop is stopped.\n\n");
	    break;
	  }
	}
        else
          lsm_quad(lsm_a, lsm_r, lsm_l, double(j+1), norf, Mp->zero,0);
      }
    }
    //////  JK

    if (j==ini || norf>err)
      {
	
	//  backup is used for actual values
	for (j=0;j<n;j++){
	  r[j]=lhsb[j];
	}
	
	//  reduction of the time increment because
	//  inner loop was not able to enforce equilibrium
	dt/=2.0;
	Mp->timecon.oldtime ();
	i--;
	
	fprintf (stdout,"\n\n time increment is reduced because the inner loop was not able to enforce equilibrium");
	if (dt<dtmin){
	  fprintf (stderr,"\n\n time increment is less than minimum time increment");
	  fprintf (stderr,"\n computation fails (file %s, line %d)\n",__FILE__,__LINE__);
          compute_req_val (lcid);
          print_step_forced(lcid, i, Mp->time, f);
          print_flush();
	  break;
	}
      }
    else
      {
	// equilibrium state was reached
	fprintf (stdout,"\n\n Last inner iteration step number j = %ld,     norf = %e \n\n",j,norf);

        //  actual total load vector becomes previous total load vector
        for (j=0;j<n;j++){
          fp[j]=f[j];
        }
	
	dtdef = Mp->timecon.actualforwtimeincr ();
	
	if (nsts==2){
	  dt*=2.0;
	  nsts=0;
	  fprintf (stdout,"\n\n time increment is enlarged because no inner loop was neccessary in previous 3 steps");
	}
	if (dt>dtdef)
	  dt=dtdef;
	  
        Mb->comp_sum (fi);
        fprintf(Out, "\n# i=%ld, time=%le[s], time=%le[d], fix=%le, fiy=%le, fiz=%le", 
                i, Mp->time, Mp->time/86400, Mb->sumcomp[0], Mb->sumcomp[1], Mb->sumcomp[2]);
	fflush(Out);
	Mm->updateipval();
	compute_req_val (lcid);
	print_step(lcid, i, Mp->time, f);
	print_flush();
	
	if ((Mp->timecon.isitimptime ()==1) && (Mp->hdbcont.hdbtype))
	  mtsolver_save (r,fp,i,Mp->time,dt,n);

	//backup of previous load
	for (j=0;j<n;j++){
	  fp[j]=f[j];
	}  
      }
  }while(Mp->time<end_time);
  
  print_close ();
  
  //if (Mp->timecon.nit > 0)
  //fclose (aux);
  
  delete [] dr;  delete [] fi;  delete [] fb;  delete [] fp;  delete [] fl;  delete [] lhsb;
}

/**
   function executes one step of time dependent mechanical algorithm
   (visco-elasticity, visco-plasticity, etc.)
   
   @param lcid - load case id
   @param f - %vector of prescribed nodal forces
   @param fi - %vector of computed nodal forces
   @param fp - %vector of previous nodal forces
   @param dr - %vector of displacement increments
   @param r - %vector of total displacements
   
   JK, 19.9.2004
*/
void one_step (long lcid,long n,double *f,double *fi,double *fp,double *fb,double *dr,double *r)
{
  long i;
  
  //  evaluation of the prescribed nodal forces at actual time
  mefel_right_hand_side (lcid,f);
  
  //  estimation of total forces in each direction
  //  used in creep
  Mb->comp_sum (f);
  
  //  nodal forces from attained stresses
  Mp->phase=1;
  internal_forces (lcid,fi);
  
  //  evaluation of the right hand side
  for (i=0;i<n;i++){
    fb[i]=f[i]-fp[i]+fi[i];
    fp[i]=f[i];
  }
  
  //  assembling of tangent stiffness matrix
  stiffness_matrix (lcid);
  
  //  solution of K(r).v=F  => increments of displacements
  //Smat->solve_system (Gtm,dr,fb);
  Mp->ssle->solve_system (Gtm,Smat,dr,fb,Out);
  
  //  new displacement vector
  for (i=0;i<n;i++){
    r[i]+=dr[i];
  }
  
  //  update of values with respect to attained state
  Mp->phase=2;
  internal_forces (lcid,fi);
  
}


void aux_nonlintime_print (FILE *aux,double *r,double l)
{
/*  long i,j,k,m,n;
  
  n=Mp->npun;
  
  fprintf (aux,"%e",l);
  for (i=0;i<n;i++){
    //  node number
    j=Mp->requn[2*i+0];
    //  unknown number
    k=Mp->requn[2*i+1];
    //  code number
    m=Mt->give_dof (j,k)-1;
    fprintf (aux,"  %e",r[m]);
  }
  fprintf (aux,"\n");
  fflush (aux);*/
}



/**
  Function saves all necessary data to backup file in the given time step. The process is driven by 
  backup controler Mp->hdbcont. The data can be used for restarting analysis from the given step.
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_save (double *r,double *fp,long ni,double time, double dt, long n)
{
  switch (Mp->hdbcont.hdbfmts)
  {
    case text:
      mtsolver_save_text(r, fp, ni, time, dt, n);
      break;
    case binary:
      mtsolver_save_binary(r, fp, ni, time, dt, n);
      break;
    default: 
      print_err("unknown type of backup file format is required", __FILE__, __LINE__, __func__);
  }
}



/**
  Function saves all necessary data to the text backup file in the given time step. The process is driven by 
  backup controler Mp->hdbcont. The data can be used for restarting analysis from the given step.
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_save_text (double *r,double *fp,long ni,double time, double dt, long n)
{
  switch (Mp->hdbcont.hdbtype)
  {
    case hdbs_single:
    case hdbrs_single:
      mtsolver_save_text_single(r, fp, ni, time, dt, n);
      break;
    case hdbs_multiple:
    case hdbrs_multiple:
      mtsolver_save_text_multiple(r, fp, ni, time, dt, n);
      break;
    default:
      print_err("unknown type of backup is required", __FILE__, __LINE__, __func__);
  }
}



/**
  Function saves all necessary data to the binary backup file in the given time step. The process is driven by 
  backup controler Mp->hdbcont. The data can be used for restarting analysis from the given step.
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_save_binary (double *r,double *fp,long ni,double time, double dt, long n)
{
  switch (Mp->hdbcont.hdbtype)
  {
    case hdbs_single:
    case hdbrs_single:
      mtsolver_save_binary_single(r, fp, ni, time, dt, n);
      break;
    case hdbs_multiple:
    case hdbrs_multiple:
      mtsolver_save_binary_multiple(r, fp, ni, time, dt, n);
      break;
    default:
      print_err("unknown type of backup is required", __FILE__, __LINE__, __func__);
  }
}



/**
  Function saves all necessary data to the single text backup file in the given time step. The process is driven by 
  backup controler Mp->hdbcont. The data can be used for restarting analysis from the given step.
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_save_text_single (double *r,double *fp,long ni,double time, double dt, long n)
{
  long i;
  FILE *aux;
  char name[FNAMELEN+20];
  char emsg[FNAMELEN+100];
  int prec = (int)Mp->hdbcont.prec;

  sprintf(name, "%s.%ld.bac",Mp->hdbcont.hdbnames, ni);
  aux = fopen(name,"wt");
  if (aux==NULL)
  {
    sprintf(emsg, "cannot open backup file %s", name);
    print_err(emsg, __FILE__, __LINE__, __func__);
    abort();
  }
  fprintf (aux,"\n\n\n\n");

  //  attained time
  fprintf (aux,"%.*le\n",prec,time);
  //  attained time increment
  fprintf (aux,"%.*le\n",prec,dt);
  //  attained number of steps
  fprintf (aux,"%ld\n",ni);
  //  number of degrees of freedom
  fprintf (aux,"%ld\n",n);

  //  left hand side vector
  for (i=0;i<n;i++){
    fprintf (aux,"%.*le\n",prec,r[i]);
  }
  
  fprintf (aux,"\n");
  
  //  vector of the right hand side from previous step
  for (i=0;i<n;i++){
    fprintf (aux,"%.*le\n",prec,fp[i]);
  }
  
  //  number of integration points in the problem
  fprintf (aux,"\n\n %ld\n",Mm->tnip);

  //  data on integration points
  Mm->save_intpoints_txt (aux, Mp->hdbcont.selelems, Mp->hdbcont.selother_s);
  
  fclose (aux);
}



/**
  Function saves all necessary data to the several text backup files in the given time step.
  Individual quantities are saved into separated files, each file contains given quantity for all
  integration points or all data needed by solver. The file names consist of name specified 
  in the Mp->hdbcont, time step id, description of the saved quantity and .bac suffix. Description
  of the quantity can be :
  solver - file contains date needed by solver
  strain - array of strains at all integration points
  stress - array of stresses at all integration points
  other  - selected components of eqother array at all integration points
  nonloc - adjacent integration points and their distances at all integration points (used only for nonlocal material models)
           These points are saved only once in the adjacip() function.
  The process is driven by the backup controler Mp->hdbcont. The data can be used for restarting 
  analysis from the given step.
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_save_text_multiple (double *r,double *fp,long ni,double time, double dt, long n)
{
  long i;
  FILE *aux;
  char name[FNAMELEN+20];
  char emsg[FNAMELEN+100];
  int prec = (int)Mp->hdbcont.prec;

  sprintf(name, "%s.solver.%ld.bac",Mp->hdbcont.hdbnames, ni);
  aux = fopen(name,"wt");
  if (aux==NULL)
  {
    sprintf(emsg, "cannot open backup file %s", name);
    print_err(emsg, __FILE__, __LINE__, __func__);
    abort();
  }
  fprintf (aux,"\n\n\n\n");

  //  attained time
  fprintf (aux,"%.*le\n",prec,time);
  //  attained time increment
  fprintf (aux,"%.*le\n",prec,dt);
  //  attained number of steps
  fprintf (aux,"%ld\n",ni);
  //  number of degrees of freedom
  fprintf (aux,"%ld\n",n);

  //  left hand side vector
  for (i=0;i<n;i++){
    fprintf (aux,"%.*le\n",prec,r[i]);
  }
  
  fprintf (aux,"\n");
  
  //  vector of the right hand side from previous step
  for (i=0;i<n;i++){
    fprintf (aux,"%.*le\n",prec,fp[i]);
  }
  
  //  number of integration points in the problem
  fprintf (aux,"\n\n %ld\n",Mm->tnip);

  fclose (aux);

  //  data on integration points
  Mm->save_intpoints_txt (ni, Mp->hdbcont.selelems, Mp->hdbcont.selother_s);
  
}



/**
  Function saves all necessary data to the single binary backup file in the given time step. The process is driven by 
  backup controler Mp->hdbcont. The data can be used for restarting analysis from the given step.
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_save_binary_single (double *r,double *fp,long ni,double time, double dt, long n)
{
  FILE *aux;
  char name[FNAMELEN+30];
  char emsg[FNAMELEN+100];

  sprintf(name, "%s.%ld.bac",Mp->hdbcont.hdbnames, ni);
  aux = fopen(name,"wb");
  if (aux==NULL)
  {
    sprintf(emsg, "cannot open backup file %s", name);
    print_err(emsg, __FILE__, __LINE__, __func__);
    abort();
  }

  //  attained time
  fwrite (&time, sizeof(time), 1, aux);
  //  attained time increment
  fwrite (&dt, sizeof(dt), 1, aux);
  //  attained number of steps
  fwrite (&ni, sizeof(ni), 1, aux);
  //  number of degrees of freedom
  fwrite (&n, sizeof(n), 1, aux);

  //  left hand side vector
  fwrite(r, sizeof(*r), n, aux);
  
  //  vector of the right hand side from previous step
  fwrite(fp, sizeof(*fp), n, aux);
  
  //  number of integration points in the problem
  fwrite (&Mm->tnip, sizeof(Mm->tnip), 1, aux);

  //  data on integration points
  Mm->save_intpoints_bin (aux, Mp->hdbcont.selelems, Mp->hdbcont.selother_s); 
  fclose (aux);
}



/**
  Function saves all necessary data to the several binary backup files in the given time step.
  Individual quantities are saved into separated files, each file contains given quantity for all
  integration points or all data needed by solver. The file names consist of name specified 
  in the Mp->hdbcont, time step id, description of the saved quantity and .bac suffix. Description
  of the quantity can be :
  solver - file contains date needed by solver
  strain - array of strains at all integration points
  stress - array of stresses at all integration points
  other  - selected components of eqother array at all integration points
  nonloc - adjacent integration points and their distances at all integration points (used only for nonlocal material models)
           These points are saved only once in the adjacip() function.
  The process is driven by the backup controler Mp->hdbcont. The data can be used for restarting 
  analysis from the given step.
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_save_binary_multiple (double *r,double *fp,long ni,double time, double dt, long n)
{
  FILE *aux;
  char name[FNAMELEN+30];
  char emsg[FNAMELEN+100];

  sprintf(name, "%s.solver.%ld.bac",Mp->hdbcont.hdbnames, ni);
  aux = fopen(name,"wb");
  if (aux==NULL)
  {
    sprintf(emsg, "cannot open backup file %s", name);
    print_err(emsg, __FILE__, __LINE__, __func__);
    abort();
  }

  //  attained time
  fwrite (&time, sizeof(time), 1, aux);
  //  attained time increment
  fwrite (&dt, sizeof(dt), 1, aux);
  //  attained number of steps
  fwrite (&ni, sizeof(ni), 1, aux);
  //  number of degrees of freedom
  fwrite (&n, sizeof(n), 1, aux);

  //  left hand side vector
  fwrite(r, sizeof(*r), n, aux);
  
  //  vector of the right hand side from previous step
  fwrite(fp, sizeof(*fp), n, aux);
  
  //  number of integration points in the problem
  fwrite (&Mm->tnip, sizeof(Mm->tnip), 1, aux);

  fclose (aux);

  //  data on integration points
  Mm->save_intpoints_bin (ni, Mp->hdbcont.selelems, Mp->hdbcont.selother_s); 
}



/**
  Function restores all necessary data from the backup file. The process is driven by 
  backup controler Mp->hdbcont. Data from solver are passed through parameters of the function. 
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_restore (double *r,double *fp,long &ni,double &time, double &dt, long &n)
{
  switch (Mp->hdbcont.hdbfmtr)
  {
    case text:
      mtsolver_restore_text(r, fp, ni, time, dt, n);
      break;
    case binary:
      mtsolver_restore_binary(r, fp, ni, time, dt, n);
      break;
    default: 
      print_err("unknown type of backup file format is required", __FILE__, __LINE__, __func__);
  }
}



/**
  Function restores all necessary data from the text backup file. The process is driven by 
  backup controler Mp->hdbcont. Data from solver are passed through parameters of the function. 
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_restore_text (double *r,double *fp,long &ni,double &time, double &dt, long &n)
{
  switch (Mp->hdbcont.hdbtype)
  {
    case hdbs_single:
    case hdbrs_single:
      mtsolver_restore_text_single(r, fp, ni, time, dt, n);
      break;
    case hdbs_multiple:
    case hdbrs_multiple:
      mtsolver_restore_text_multiple(r, fp, ni, time, dt, n);
      break;
    default:
      print_err("unknown type of backup is required", __FILE__, __LINE__, __func__);
  }
}



/**
  Function restores all necessary data from the binary backup file. The process is driven by 
  backup controler Mp->hdbcont. Data from solver are passed through parameters of the function. 
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_restore_binary (double *r,double *fp,long &ni,double &time, double &dt, long &n)
{
  switch (Mp->hdbcont.hdbtype)
  {
    case hdbs_single:
    case hdbrs_single:
      mtsolver_restore_binary_single(r, fp, ni, time, dt, n);
      break;
    case hdbs_multiple:
    case hdbrs_multiple:
      mtsolver_restore_binary_multiple(r, fp, ni, time, dt, n);
      break;
    default:
      print_err("unknown type of backup is required", __FILE__, __LINE__, __func__);
  }
}



/**
  Function restores all necessary data from the one huge text backup file. The process is driven by 
  backup controler Mp->hdbcont. Data from solver are passed through parameters of the function. 
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_restore_text_single (double *r, double *fp, long &ni, double &time, double &dt, long &n)
{
  long i;
  FILE *aux;
  char name[FNAMELEN+20];
  char emsg[FNAMELEN+100];

  sprintf(name, "%s.%ld.bac",Mp->hdbcont.hdbnamer, ni);
  aux = fopen(name,"rt");
  if (aux==NULL)
  {
    sprintf(emsg, "cannot open backup file %s", name);
    print_err(emsg, __FILE__, __LINE__, __func__);
    abort();
  }
  
  //  attained time
  fscanf (aux,"%le",&time);
  //  attained number of steps
  fscanf (aux,"%ld\n",&ni);
  //  number of degrees of freedom
  fscanf (aux,"%ld\n",&n);

  if (n!=Ndofm){
    print_err("number of DOFs in backup file and in the actual problem is not same\n", __FILE__, __LINE__, __func__);
    abort();
  }

  //  left hand side vector
  for (i=0;i<n;i++){
    fscanf (aux,"%le",r+i);
  }
  //  vector of the right hand side from previous step
  for (i=0;i<n;i++){
    fscanf (aux,"%le",fp+i);
  }
  
  //  number of integration points in the problem
  fscanf (aux,"%ld",&i);
  
  if (i!=Mm->tnip){
    print_err("number of integration points in backup file and in the actual problem is not same\n", __FILE__, __LINE__, __func__);
    abort();
  }

  //  data on integration points
  Mm->restore_intpoints_txt (aux, Mp->hdbcont.selelemr, Mp->hdbcont.selother_r, Mp->hdbcont.selother_id);

  fclose(aux);
}



/**
  Function restores all necessary data from the separated text backup file. The process is driven by 
  backup controler Mp->hdbcont. Data from solver are passed through parameters of the function. 
  Individual quantities have to be saved into separated files, each file contains given quantity for all
  integration points or all data needed by solver. The file names consist of name specified 
  in the Mp->hdbcont, time step id, description of the saved quantity and .bac suffix. Description
  of the quantity can be :
  solver - file contains date needed by solver
  strain - array of strains at all integration points
  stress - array of stresses at all integration points
  other  - selected components of eqother array at all integration points
  nonloc - adjacent integration points and their distances at all integration points (used only for nonlocal material models)
           These points are saved only once in the adjacip() function.
  The process is driven by the backup controler Mp->hdbcont. The data can be used for restarting 
  analysis from the given step.
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_restore_text_multiple (double *r,double *fp,long &ni,double &time, double &dt, long &n)
{
  long i;
  FILE *aux;
  char name[FNAMELEN+20];
  char emsg[FNAMELEN+100];

  sprintf(name, "%s.solver.%ld.bac",Mp->hdbcont.hdbnamer, ni);
  aux = fopen(name,"rt");
  if (aux==NULL)
  {
    sprintf(emsg, "cannot open backup file %s", name);
    print_err(emsg, __FILE__, __LINE__, __func__);
    abort();
  }
  
  //  attained time
  fscanf (aux,"%le",&time);
  //  attained number of steps
  fscanf (aux,"%ld\n",&ni);
  //  number of degrees of freedom
  fscanf (aux,"%ld\n",&n);

  if (n!=Ndofm){
    print_err("number of DOFs in backup file and in the actual problem is not same\n", __FILE__, __LINE__, __func__);
    abort();
  }

  //  left hand side vector
  for (i=0;i<n;i++){
    fscanf (aux,"%le",r+i);
  }
  //  vector of the right hand side from previous step
  for (i=0;i<n;i++){
    fscanf (aux,"%le",fp+i);
  }
  
  //  number of integration points in the problem
  fscanf (aux,"%ld",&i);
  
  if (i!=Mm->tnip){
    print_err("number of integration points in backup file and in the actual problem is not same\n", __FILE__, __LINE__, __func__);
    abort();
  }

  fclose(aux);

  //  data on integration points
  Mm->restore_intpoints_txt (Mp->hdbcont.selelemr, Mp->hdbcont.selother_r, Mp->hdbcont.selother_id);
}



/**
  Function restores all necessary data from the one huge binary backup file. The process is driven by 
  backup controler Mp->hdbcont. Data from solver are passed through parameters of the function. 
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_restore_binary_single (double *r,double *fp,long &ni,double &time, double &dt, long &n)
{
  FILE *aux;
  char name[FNAMELEN+30];
  char emsg[FNAMELEN+100];

  sprintf(name, "%s.%ld.bac",Mp->hdbcont.hdbnamer, ni);
  aux = fopen(name,"rb");
  if (aux==NULL)
  {
    sprintf(emsg, "cannot open backup file %s", name);
    print_err(emsg, __FILE__, __LINE__, __func__);
    abort();
  }

  //  attained time
  fread (&time, sizeof(time), 1, aux);
  //  attained time increment
  fread (&dt, sizeof(dt), 1, aux);
  //  attained number of steps
  fread (&ni, sizeof(ni), 1, aux);
  //  number of degrees of freedom
  fread (&n, sizeof(n), 1, aux);

  //  left hand side vector
  fread(r, sizeof(*r), n, aux);
  
  //  vector of the right hand side from previous step
  fread(fp, sizeof(*fp), n, aux);
  
  //  number of integration points in the problem
  fread (&Mm->tnip, sizeof(Mm->tnip), 1, aux);

  //  data on integration points
  Mm->restore_intpoints_bin (aux, Mp->hdbcont.selelemr, Mp->hdbcont.selother_r, Mp->hdbcont.selother_id);
  fclose (aux);
}



/**
  Function restores all necessary data from the separated text backup file. The process is driven by 
  backup controler Mp->hdbcont. Data from solver are passed through parameters of the function. 
  Individual quantities have to be saved into separated files, each file contains given quantity for all
  integration points or all data needed by solver. The file names consist of name specified 
  in the Mp->hdbcont, time step id, description of the saved quantity and .bac suffix. Description
  of the quantity can be :
  solver - file contains date needed by solver
  strain - array of strains at all integration points
  stress - array of stresses at all integration points
  other  - selected components of eqother array at all integration points
  nonloc - adjacent integration points and their distances at all integration points (used only for nonlocal material models)
           These points are saved only once in the adjacip() function.
  The process is driven by the backup controler Mp->hdbcont. The data can be used for restarting 
  analysis from the given step.
  
  @param r - %vector of unknowns
  @param fp - %vector of actual right hand side
  @param ni - time step id
  @param time - actual time of saved step
  @param dt   - actual time increment of saved step
  @param n    - number of unknowns (size of %vectors r and fp)

  9.2008 TKo  
*/
void mtsolver_restore_binary_multiple (double *r,double *fp,long &ni,double &time, double &dt, long &n)
{
  FILE *aux;
  char name[FNAMELEN+30];
  char emsg[FNAMELEN+100];

  sprintf(name, "%s.solver.%ld.bac",Mp->hdbcont.hdbnamer, ni);
  aux = fopen(name,"wb");
  if (aux==NULL)
  {
    sprintf(emsg, "cannot open backup file %s", name);
    print_err(emsg, __FILE__, __LINE__, __func__);
    abort();
  }

  //  attained time
  fread (&time, sizeof(time), 1, aux);
  //  attained time increment
  fread (&dt, sizeof(dt), 1, aux);
  //  attained number of steps
  fread (&ni, sizeof(ni), 1, aux);
  //  number of degrees of freedom
  fread (&n, sizeof(n), 1, aux);

  //  left hand side vector
  fread(r, sizeof(*r), n, aux);
  
  //  vector of the right hand side from previous step
  fread(fp, sizeof(*fp), n, aux);
  
  //  number of integration points in the problem
  fread (&Mm->tnip, sizeof(Mm->tnip), 1, aux);

  fclose (aux);

  //  data on integration points
  Mm->restore_intpoints_bin (Mp->hdbcont.selelemr, Mp->hdbcont.selother_r, Mp->hdbcont.selother_id);
}



