#include "ortodam.h"
#include "bmatrix.h"
#include "matrix.h"
#include "vector.h"
#include "elastisomat.h"
#include "global.h"
#include "intpoints.h"
#include "tensor.h"
#include "vecttens.h"
#include <math.h>
#include <stdlib.h>

#define nijac 200
#define limit 1.0e-15
#ifndef M_PI
 #define M_PI 3.14159265358979323846
#endif





/**
  This constructor inializes attributes to zero values.
  
  TKo 9.2006
  modified TKo 10.2008
*/
ortodam::ortodam (void)
{
  cde = corr_off;
  ac = 0.0;  bc = 0.0;  y0c = 0.0;  at = 0.0;  bt = 0.0;  y0t = 0.0;
  gfc = 0.0; gft = 0.0;
  ft = uft = 0.0;
  fc = ufc = 0.0;
}

/**
  This destructor is only for the formal purposes.
  
  TKo 9.2006
*/
ortodam::~ortodam (void)
{

}

/**
  This function reads material parameters from the opened text file given
  by the parameter in.

  @param in - pointer to the opened input text file
  
  TKo 9.2006
  modified TKo 10.2008
*/
void ortodam ::read (XFILE *in)
{
  xfscanf (in, "%k%m", "dam_evol_f", &dam_evolfunc_kwdset, (int *)&damevf);
  switch(damevf)
  {
    case brittle:
      xfscanf (in,"%k%lf %k%lf", "f_t", &ft, "u_ft", &uft);
      xfscanf (in,"%k%lf %k%lf", "f_c", &fc, "u_fc", &ufc);
      xfscanf (in,"%k%m", "corr_dissip", &corr_disip_en_kwdset, (int *)(&cde));
      if (cde != corr_off)
        sra.read (in);
      break;
    case quasi_brittle:
      xfscanf (in, "%k%m", "corr_dissip", &corr_disip_en_kwdset, (int *)(&cde));
      switch (cde)
      {
        case corr_off:
          xfscanf (in, "%k%lf %k%lf %k%lf", "A_c", &ac, "B_c", &bc, "Y_0c", &y0c);
          xfscanf (in, "%k%lf %k%lf %k%lf", "A_t", &at, "B_t", &bt, "Y_0t", &y0t);
          break;
        case corr_on:
          xfscanf (in, "%k%lf %k%lf %k%lf", "G_fc", &gfc, "B_c", &bc, "Y_0c", &y0c);
          xfscanf (in, "%k%lf %k%lf %k%lf", "G_ft", &gft, "B_t", &bt, "Y_0t", &y0t);
          break;
        default:
          print_err(" unknown type of correction of disipated energy is required", __FILE__, __LINE__,__func__);
      }
      break;
    default:
      print_err(" unknown type of damage evolution function is required", __FILE__, __LINE__,__func__);
  }
}



/**
  Function computes value of load function 

  @param  ipp  - integration point number
  @param  nu   - Poissons ratio
  @param  peps - %vector of principal strains
  @param  damt - %vector actual values of principal damage parameters for tension
  @param  damc - %vector actual values of principal damage parameters for compression
  @param  aat  - actual values of material parameter a for tension 
  @param  aac  - actual values of material parameter a for compression 
  @param  lft  - %vector of resulting values of load function for tension for each principal direction
  @param  lfc  - %vector of resulting values of load function for compression for each principal direction

  @retval The function returns load function values separately for tension and compression
          via parameters lft and lfc for each principal direction.

  TKo 9.2006
  modified TKo 10.2008
*/
void ortodam::loadfunc(long ipp, double nu, vector &peps, vector &damt, vector &damc, 
                       double aat, double aac, vector &lft, vector &lfc)
{
  long i;
  double e, af, auf, kappa, tmp1, tmp2;
  vector pyt(3), pyc(3);

  switch(damevf)
  {
    case brittle:
      for (i=0; i<3; i++)
      {
        // for tension
        if (peps(i) > 0.0)
	{
          kappa = peps(i);
          if (kappa <= y0t) 
            lft(i) = -1.0;      
          else
          {
            e = Mm->give_actual_ym(ipp);
            // actual tensile strength
            af = Mm->give_actual_ft(ipp);
            // actual value of initial crack opening
            auf = uft/(af/ft);
            tmp1 = -(kappa-af/e)/(auf-af/e);
            lft(i) = 1.0 - af/(e*kappa)*exp(tmp1) - damt(i);
	  }
	}
        // for compression
        if (peps(i) < 0.0)
	{
	  kappa = fabs(peps(i));
          if (kappa <= y0c) 
            lfc(i) = -1.0;      
          else
          {
            e = Mm->give_actual_ym(ipp);
            // actual compressive strength
            af = Mm->give_actual_fc(ipp);
            // actual value of initial crack opening
            auf = ufc/(af/fc);
            tmp1 = -(kappa-af/e)/(auf-af/e);
            lfc(i) = 1.0 - af/(e*kappa)*exp(tmp1) - damc(i);
	  }
        }
      }
      break;
    case quasi_brittle:
      for (i=0; i<3; i++)
      {
        // for tension
        if (peps(i) > 0.0)
	{
          kappa = peps(i);
          if (kappa <= y0t) 
            lft(i) = -1.0;      
          else
          {
            // compute (yt-y0t)^bt
            tmp1 = kappa - y0t;
            tmp2 = log(tmp1);
            tmp1 = exp(bt*tmp2);
            // compute value of load function for tension in given principal direction i
            lft(i) = (1.0 - damt(i))*(1.0+aat*tmp1) - 1.0;
          }
	}
        // for compression
        if (peps(i) < 0.0)
	{
          kappa = fabs(peps(i));
          if (kappa <= y0c) 
            lfc(i) = -1.0;      
          else
          {
            // compute (yc-y0c)^bc
            tmp1 = kappa - y0c;
            tmp2 = log(tmp1);
            tmp1 = exp(bc*tmp2);
            // compute value of load function for compression in given principal direction i
            lfc(i) = (1.0 - damc(i))*(1.0+aac*tmp1) - 1.0;
          }
	}
      }
      break;
    default:
      print_err(" unknown type of damage evolution function is required", __FILE__, __LINE__,__func__);
  }  
  return;
}


/**
  The function computes damage parameter which is the result of the
  damage function for brittle type of damage function. Correction of dissipated
  energy is controled by the cde flag.

  @param ipp - integration point number
  @param y -  parameter of damage function=equivalent strain (generalized conjugated dimensionless thermodynamical force)
  @param e - actual value of Young modulus
  @param f - actual value of strength (tensile or compressive)
  @param uf - actual value of initial crack opening
  
  @retval Function returns value of the damage parameter omega.
  
  TKo
*/
double ortodam::brittle_damage(long ipp, double y, double e, double f, double uf)
{
  double err,omega = 0.0, dtmp, tmp, h, indam, kappa;
  long i,ni, eid;
  
  ni=sra.give_ni ();
  err=sra.give_err ();
  
  indam = f/e;
  
  kappa = y;
  if (kappa < indam)
    // elastic loading
    return 0.0;
  if ((Mm->ip[ipp].hmt & 2) > 0)
  {
    omega = 1.0-(f/(e*kappa))*exp(-(kappa-f/e)/(uf-f/e));
    return omega;
  }
  // correction of dispiated energy - mesh independent scalar damage
  eid = Mm->elip[ipp];
  switch (Mt->give_dimension(eid))
    {
    case 1 :
      h = Mt->give_length(eid);
      break;
    case 2 :
      h = sqrt(Mt->give_area(eid));
      break;
    case 3 :
      h = pow(Mt->give_volume(eid), 1.0/3.0);
      break;
    default :
      print_err("unknown dimension of element", __FILE__, __LINE__, __func__);
    }
  if (cde == corr_off)
  // no correction of disipated energy only simple scalar damge
  { 
    omega = 1.0-(f/(e*kappa))*exp(-h*(kappa-f/e)/(uf-f/e));
    return omega;
  }
  for (i = 0; i < ni; i++)
  {
    dtmp = -e*kappa+f/uf*h*kappa*exp(-h*omega*kappa/uf);
    tmp = (1.0-omega)*e*kappa-f*exp(-h*omega*kappa/uf);
    if (fabs(tmp) < err)
    {
      omega += - tmp/dtmp;
      break;
    }
    omega += - tmp/dtmp;
  }
  if (fabs(tmp) > err)
    print_err("iteration of omega doesn't converge with required error", __FILE__, __LINE__, __func__);
  
  if ((omega > 1.0) || (omega < 0.0))
    print_err("iteration of omega doesn't converge to range <0,1>", __FILE__, __LINE__, __func__);
  
  return omega;
}



/**
  Function computes values of principal damage parameters Dt and Dc. The damage parameters
  are derived from the damage function.

  @param ipp  - integration point number
  @param peps - %vector of principal strains
  @param aat  - actual value of material parameter at 
  @param aac  - actual value of material parameter ac 
  @param pdamc - %vector of principal damage parameters for tension
  @param pdamt - %vector of principal damage parameters for compression
  
  TKo 9.2006
*/
void ortodam::princ_dam(long ipp, vector &peps, double aat, double aac, vector &pdamt, vector &pdamc)
{
  long i;
  double tmp, ltmp;
  double e, af, auf;

  fillv(0.0, pdamt);
  fillv(0.0, pdamc);

  switch(damevf)
  {
    case brittle:
      e = Mm->give_actual_ym(ipp);
      for(i=0; i<3; i++)
      {           
        // tension 
        if (peps(i) > 0.0)
	{
          // actual tensile strength
          af = Mm->give_actual_ft(ipp);
          // actual value of initial crack opening
          auf = uft/(af/ft);
          pdamt(i) = brittle_damage(ipp, peps(i), e, af, auf);
	}
        // compression 
        if (peps(i) < 0.0)
	{
          // actual compressive strength
          af = Mm->give_actual_fc(ipp);
          // actual value of initial crack opening
          auf = ufc/(af/fc);
          pdamc(i) = brittle_damage(ipp, fabs(peps(i)), e, af, auf);
	}
      }
      break;
    case quasi_brittle:
      for(i=0; i<3; i++)
      {    
        // tension 
        if (peps(i) > 0.0)
	{
          if (peps(i) > y0t)
          {    
            ltmp = log(peps(i) - y0t);
            pdamt(i) = aat*exp(ltmp*bt);
            tmp = 1.0/(1.0+pdamt(i));
            pdamt(i) = 1.0 - tmp;
          }
          if (pdamt(i) < 0.0)
          {
            fprintf(stdout, "\nDamage < 0.0 (ipp=%ld)\n", ipp);
            abort();
          }
	}
        // compression 
        if (peps(i) < 0.0)
	{
          if (fabs(peps(i)) > y0c)
          {    
            ltmp = log(fabs(peps(i)) - y0c);
            pdamc(i) = aac*exp(ltmp*bc);
            tmp = 1.0/(1.0+pdamc(i));
            pdamc(i) = 1.0-tmp;
          }
          if (pdamc(i) < 0.0)
          {
            fprintf(stdout, "\nDamage < 0.0 (ipp=%ld)\n", ipp);
            abort();
          }
	}
      }
      break;
    default:
      print_err(" unknown type of damage evolution function is required", __FILE__, __LINE__,__func__);
  }
}



/**
  Function computes material parameters A, At and Ac. 
  If the correction of dissipated energy is required, they are computed with respect 
  to size of element for given ipp and variable softening modulus technique was used in 
  order to avoid mesh dependence. Otherwise, values read from the input file are used.

  @param d - allocated matrix structure for material stiffness %matrix
  @param ipp - integration point number
  @param ido - index of internal variables for given material in the ipp other array
  @param aat - actual value of parameter At 
  @param aac - actual value of parameter Ac 

  @retval The function returns computed 
  
  TKo 9.2006
*/
void ortodam::give_actual_param_a(long ipp, long ido, double &aat, double &aac)
{
  long eid;
  double e, h, tmp;

  aat = Mm->ip[ipp].eqother[ido];
  aac = Mm->ip[ipp].eqother[ido+1];
  if ((aac == 0.0) || (aat == 0.0))
  {
    if (cde == corr_off)
    // no correction of dissipated energy is required
    {
      aat = at;
      aac = ac;
      Mm->ip[ipp].eqother[ido] = aat;
      Mm->ip[ipp].eqother[ido+1] = aac;
      return;
    }
    eid = Mm->elip[ipp];
    switch (Mt->give_dimension(eid))
    {
      case 1 :
        h = Mt->give_length(eid);
        break;
      case 2 :
        h = sqrt(Mt->give_area(eid));
        break;
      case 3 :
        h = pow(Mt->give_volume(eid), 1.0/3.0);
        break;
      default :
        fprintf(stderr, "\nError determining dimension of element");
        fprintf(stderr, " in function ortodam::give_actual_param_a\n(file %s, line %d, elem : %ld)\n", __FILE__, __LINE__, eid+1);
    }
    e = Mm->give_actual_ym(ipp);
    // variable softening for tension
    tmp = (gft/h - y0t/e)*bt*e*sin(M_PI/bt)/M_PI; 
    if (tmp < 0.0)
    {
      fprintf(stderr, "\nError while computing variable softening for tension");
      fprintf(stderr, " in function ortodam::give_actual_param_a\n(file %s, line %d, elem : %ld)\n", __FILE__, __LINE__, eid+1);     
    }
    aat = pow(tmp, -bt);
    // variable softening for compression
    tmp = (gfc/h - y0c/e)*bc*e*sin(M_PI/bc)/M_PI; 
    if (tmp < 0.0)
    {
      fprintf(stderr, "\nError while computing variable softening for compression");
      fprintf(stderr, " in function ortodam::give_actual_param_a\n(file %s, line : %d, elem : %ld)\n", __FILE__, __LINE__, eid+1);     
    }
    aac = pow(tmp, -bc);
    Mm->ip[ipp].eqother[ido] = aat;
    Mm->ip[ipp].eqother[ido+1] = aac;
  }
}



/**
  This function returns the value of tensile strength

  @param ipp - integration point number in the mechmat ip array.
  @param ido - index of internal variables for given material in the ipp other array
  
  10.2008 TKo
*/
double ortodam::give_actual_ft(long ipp)
{
  return ft;
}



/**
  This function returns the value of compressive strength

  @param ipp - integration point number in the mechmat ip array.
  @param ido - index of internal variables for given material in the ipp other array
  
  10.2008 TKo
*/
double ortodam::give_actual_fc(long ipp)
{
  return fc;
}



/**
  Function computes initializes material parameters A, At and Ac. 

  @param d - allocated matrix structure for material stiffness %matrix
  @param ipp - integration point number
  @param ido - index of internal variables for given material in the ipp other array

  @retval The function returns computed 
  
  TKo 9.2006
*/
void ortodam::initvalues(long ipp, long ido)
{
  double aat, aac;
  give_actual_param_a(ipp, ido, aac, aat);
}



/**
  This function computes material stiffnes %matrix.

  @param d - allocated matrix structure for material stiffness %matrix
  @param ipp - integration point number
  @param ido - index of internal variables for given material in the ipp other array
  
  TKo
*/
void ortodam::matstiff (matrix &d,long ipp,long ido)
{
//  long ncompstr = Mm->ip[ipp].ncompstr;

  switch (Mp->stmat)
  {
    case initial_stiff :
      elmatstiff(d, ipp);
      break;
    case tangent_stiff :
      elmatstiff(d, ipp);
//      tmatstiff(d, ipp, ido);
      break;
    default :
      fprintf(stderr, "\n\nError - unknown type of stifness matrix");
      fprintf(stderr, "\n in function ortodam::matstiff (file %s, line %d)\n", __FILE__, __LINE__);
  }
}



/**
  This function computes elastic material stiffnes %matrix from actual Young modulus.

  @param d - allocated matrix structure for material stiffness %matrix
  @param ipp - integration point number

  TKo
*/
void ortodam::elmatstiff (matrix &d,long ipp)
{
  double e, e0;

  Mm->elmatstiff (d,ipp);
  e = Mm->give_actual_ym(ipp);
  e0 = Mm->give_initial_ym(ipp);
  cmulm(e/e0, d);
}




/**
  This function computes material stiffnes %matrix used in principal stress computations.

  @param d - allocated matrix structure for material stiffness %matrix
  @param e - actual value of Young modulus
  @param nu - actual value of Poissons ratio
  
  TKo
*/
void ortodam::pelmatstiff (long ipp, matrix &d, double e, double nu)
{
  double b;

  switch (Mm->ip[ipp].ssst)
  {
    case planestress :
      b = e/(1.0-nu*nu);
      d[0][0] = b;     d[0][1] = b*nu;  d[0][2] = b*nu;
      d[1][0] = b*nu;  d[1][1] = b;     d[1][2] = b*nu;
      d[2][0] = b*nu;  d[2][1] = b*nu;  d[2][2] = b;
      break;
    case planestrain :
    case spacestress :
      b = e /((1.0 + nu)*(1.0 - 2.0*nu));
      d[0][0] = d[1][1] = d[2][2] = 1.0 - nu;
      d[0][1] = d[0][2] = d[1][0] = d[1][2] = d[2][0] = d[2][1] = nu;
      cmulm(b, d);
      break;
    default :
      print_err("Unknown type of stress/strain state is required", __FILE__, __LINE__, __func__);
      break;
  }
}



/**
  This function computes correct stresses in the integration point and stores
  them into ip stress array.

  @param ipp - integration point pointer
  @param im  - index of material type for given ip
  @param ido - index of internal variables for given material in the ipp other array

  TKo, 9.2006
*/
void ortodam::nlstresses (long ipp, long im, long ido)
{
  long i;
  long ncompstr=Mm->ip[ipp].ncompstr;

  // actual material parameters a 
  double aat, aac;
  // old values of driving forces
  //double yo;
  vector pyto(3), pyco(3);
  // old values of damage parameters
  vector pdto(3), pdco(3);
  // damage parameters
  vector pdt(3), pdc(3);
  // strains and principal strains
  vector epsn(ncompstr);
  // tensor form of strains and transformation matrix for principal direction of strains
  matrix epst(3,3), t(3,3);
  // stress tensor and stress vector
  matrix sigt(3,3), d(3,3);
  vector sigma(ncompstr);
  // bulk modulus*3, shear modulus, volumetric strain, Young modulus, Poissons ratio
  double e, nu;
  // actual values of load fuctions
  //double lf;
  vector peps(3), psig(3), lft(3), lfc(3),pepsp(3);
  double h;
  long eid;
  // coefficients for inelastic strains evolution
  // double betat = 1.0e-4;
  // double betac = 1.0e-4;
  // double gamma = 1.0e-4;

  // initialize values of material parameter a
  give_actual_param_a(ipp, ido, aat, aac);
  e  = Mm->give_actual_ym(ipp);
  nu = Mm->give_actual_nu(ipp);
  if (Mm->ip[ipp].ssst == planestress)
    Mm->ip[ipp].strain[3] = -nu / (1.0 - nu) * (Mm->ip[ipp].strain[0]+Mm->ip[ipp].strain[1]);

  //  total new strains
  for (i=0;i<ncompstr;i++)
    epsn[i] = Mm->ip[ipp].strain[i];

  // previous principal damage parameters
  for (i=0; i<3; i++)
  {
    // damage parameters
    pdto[i] = Mm->ip[ipp].eqother[ido+2+i];   // tension
    pdco[i] = Mm->ip[ipp].eqother[ido+2+3+i]; // compression    
  }
  
  // new driving forces
  vector_tensor(epsn, epst, Mm->ip[ipp].ssst, strain);
  princ_val (epst,peps,t,nijac,limit,Mp->zero,3, 1);
  
  /* for (i=0; i<3; i++)
  {
    for(j=0; j<3; j++)
      Mm->ip[ipp].other[ido+3+1+3+3+1+3+3+3*i+j] = t[i][j]; // 
  }  
  // inelastic strains due to previous damage
  for (i=0; i<3; i++)
  {
    pepsp[i]  = betat*pdto[i]*pdto[i]/(1.0-pdto[i]);
    pepsp[i] -= betac*pdco[i]*pdco[i]/(1.0-pdco[i]);
    pepsp[i] -= gamma*dvo*dvo/(1.0-dvo);
  }
  subv(peps,pepsp,peps);*/

  princ_dam(ipp, peps, aat, aac, pdt, pdc);
  for (i=0; i<3; i++)
  {
    if (pdt[i] < pdto[i])
      pdt[i] = pdto[i];
    if (pdc[i] < pdco[i])
      pdc[i] = pdco[i];
  }

  // new principal stresses
  pelmatstiff (ipp, d, e, nu);
  mxv(d, peps, psig);
  for (i=0; i<3; i++)
  {
    if (psig(i) > 0.0)
      psig(i) *= 1.0-pdt(i);
    if (psig(i) < 0.0)
      psig(i) *= 1.0-pdc(i);
  }
   
  // transformation of principal stresses to the global coordinate system
  fillm(0.0, sigt);
  for (i = 0; i < 3; i++)
    sigt[i][i] = psig[i];
  lgmatrixtransf(sigt, t);
  tensor_vector(sigma, sigt, Mm->ip[ipp].ssst, stress);
  if (Mm->ip[ipp].ssst == planestress)
    sigma[3] = 0.0;

  //  new data storage
  for (i=0;i<ncompstr;i++)
    Mm->ip[ipp].stress[i]=sigma[i];
  for (i=0; i<3; i++)
  {
    // damage parameters
    Mm->ip[ipp].other[ido+2+i]   = pdt[i]; // tension
    Mm->ip[ipp].other[ido+2+3+i] = pdc[i]; // compression    
  }
  // following is only for debugging purposes
  // storage of tensile damage deviatoric part of tensor
  fillm(0.0, sigt);
  for (i = 0; i < 3; i++)
    sigt[i][i] = pdt[i];
  lgmatrixtransf(sigt, t);
  for (i = 0; i < 3; i++)
    Mm->ip[ipp].other[ido+9+i] = sigt[i][i];
  Mm->ip[ipp].other[ido+12] = sigt[1][2];       
  Mm->ip[ipp].other[ido+13] = sigt[0][2];       
  Mm->ip[ipp].other[ido+14] = sigt[0][1];       
  // computation of approximate value of crack opening
  eid = Mm->elip[ipp];
  switch (Mt->give_dimension(eid))
  {
    case 1 :
      h = Mt->give_length(eid);
      break;
    case 2 :
      h = sqrt(Mt->give_area(eid));
      break;
    case 3 :
      h = pow(Mt->give_volume(eid), 1.0/3.0);
      break;
    default :
      print_err("unknown dimension of element", __FILE__, __LINE__, __func__);
  }
  // storage of cracks openings computed from tensile damage tensor
  for (i = 0; i < 3; i++)
  {
    if (peps(i) > 0.0)
      Mm->ip[ipp].other[ido+15+i] = pdt[i]*(sqrt(peps[i])-psig[i]/e)*h;
    else
      Mm->ip[ipp].other[ido+15+i] = 0.0;
  }

   
}

/**
  Function updates values in the other array reached in the previous equlibrium state to
  values reached in the new actual equilibrium state.

  @param ipp - integration point number in the mechmat ip array.
  @param im  - index of material type for given ip
  @param ido - index of internal variables for given material in the ipp other array
  
  TKo
*/
void ortodam::updateval (long ipp,long im,long ido)
{
  long i;
  long n = Mm->givencompeqother(ipp,im);

  for (i=0;i<n;i++)
    Mm->ip[ipp].eqother[ido+i]=Mm->ip[ipp].other[ido+i];
}
